@@ -39,10 +39,6 @@
 
 #include "config.h"
 
-#if RANDR_MAJOR > 1 || (RANDR_MAJOR == 1 && RANDR_MINOR >= 2)
-#define HAS_RANDR_1_2 1
-#endif
-
 static char	*program_name;
 static Display	*dpy;
 static Window	root;
@@ -111,15 +107,10 @@ usage(void)
     fprintf(stderr, "  --verbose\n");
     fprintf(stderr, "  --dryrun\n");
     fprintf(stderr, "  --nograb\n");
-#if HAS_RANDR_1_2
     fprintf(stderr, "  --prop or --properties\n");
     fprintf(stderr, "  --fb <width>x<height>\n");
     fprintf(stderr, "  --fbmm <width>x<height>\n");
     fprintf(stderr, "  --dpi <dpi>/<output>\n");
-#if 0
-    fprintf(stderr, "  --clone\n");
-    fprintf(stderr, "  --extend\n");
-#endif
     fprintf(stderr, "  --output <output>\n");
     fprintf(stderr, "      --auto\n");
     fprintf(stderr, "      --mode <mode>\n");
@@ -149,7 +140,6 @@ usage(void)
     fprintf(stderr, "  --rmmode <name>\n");
     fprintf(stderr, "  --addmode <output> <name>\n");
     fprintf(stderr, "  --delmode <output> <name>\n");
-#endif
 
     exit(1);
     /*NOTREACHED*/
@@ -179,6 +169,12 @@ warning (const char *format, ...)
     va_end (ap);
 }
 
+/* Because fmin requires C99 suppport */
+static inline double dmin (double x, double y)
+{
+    return x < y ? x : y;
+}
+
 static char *
 rotation_name (Rotation rotation)
 {
@@ -209,13 +205,12 @@ reflection_name (Rotation rotation)
     return "invalid reflection";
 }
 
-#if HAS_RANDR_1_2
-typedef enum _policy {
-    clone, extend
-} policy_t;
-
 typedef enum _relation {
-    left_of, right_of, above, below, same_as,
+    relation_left_of,
+    relation_right_of,
+    relation_above,
+    relation_below,
+    relation_same_as,
 } relation_t;
 
 typedef struct {
@@ -311,7 +306,7 @@ struct _output {
     crtc_t	    *current_crtc_info;
     
     name_t	    mode;
-    float	    refresh;
+    double	    refresh;
     XRRModeInfo	    *mode_info;
     
     name_t	    addmode;
@@ -333,7 +328,11 @@ struct _output {
 	float blue;
     } gamma;
 
+    float	    brightness;
+
     Bool	    primary;
+
+    Bool	    found;
 };
 
 typedef enum _umode_action {
@@ -376,7 +375,7 @@ static int	num_crtcs;
 static XRRScreenResources  *res;
 static int	fb_width = 0, fb_height = 0;
 static int	fb_width_mm = 0, fb_height_mm = 0;
-static float	dpi = 0;
+static double	dpi = 0;
 static char	*dpi_output = NULL;
 static Bool	dryrun = False;
 static int	minWidth, maxWidth, minHeight, maxHeight;
@@ -489,27 +488,27 @@ mode_geometry (XRRModeInfo *mode_info, Rotation rotation,
 }
 
 /* v refresh frequency in Hz */
-static float
+static double
 mode_refresh (XRRModeInfo *mode_info)
 {
-    float rate;
+    double rate;
     
     if (mode_info->hTotal && mode_info->vTotal)
-	rate = ((float) mode_info->dotClock / 
-		((float) mode_info->hTotal * (float) mode_info->vTotal));
+	rate = ((double) mode_info->dotClock /
+		((double) mode_info->hTotal * (double) mode_info->vTotal));
     else
     	rate = 0;
     return rate;
 }
 
 /* h sync frequency in Hz */
-static float
+static double
 mode_hsync (XRRModeInfo *mode_info)
 {
-    float rate;
+    double rate;
     
     if (mode_info->hTotal)
-	rate = (float) mode_info->dotClock / (float) mode_info->hTotal;
+	rate = (double) mode_info->dotClock / (double) mode_info->hTotal;
     else
     	rate = 0;
     return rate;
@@ -632,6 +631,8 @@ add_output (void)
     if (!output)
 	fatal ("out of memory\n");
     output->next = NULL;
+    output->found = False;
+    output->brightness = 1.0;
     *outputs_tail = output;
     outputs_tail = &output->next;
     return output;
@@ -711,11 +712,11 @@ find_crtc_by_xid (RRCrtc crtc)
 }
 
 static XRRModeInfo *
-find_mode (name_t *name, float refresh)
+find_mode (name_t *name, double refresh)
 {
     int		m;
     XRRModeInfo	*best = NULL;
-    float	bestDist = 0;
+    double	bestDist = 0;
 
     for (m = 0; m < res->nmode; m++)
     {
@@ -727,7 +728,7 @@ find_mode (name_t *name, float refresh)
 	}
 	if ((name->kind & name_string) && !strcmp (name->string, mode->name))
 	{
-	    float   dist;
+	    double   dist;
 	    
 	    if (refresh)
 		dist = fabs (mode_refresh (mode) - refresh);
@@ -772,7 +773,7 @@ find_mode_for_output (output_t *output, name_t *name)
     XRROutputInfo   *output_info = output->output_info;
     int		    m;
     XRRModeInfo	    *best = NULL;
-    float	    bestDist = 0;
+    double	    bestDist = 0;
 
     for (m = 0; m < output_info->nmode; m++)
     {
@@ -787,7 +788,7 @@ find_mode_for_output (output_t *output, name_t *name)
 	}
 	if ((name->kind & name_string) && !strcmp (name->string, mode->name))
 	{
-	    float   dist;
+	    double   dist;
 
 	    /* Stay away from doublescan modes unless refresh rate is specified. */
 	    if (!output->refresh && (mode->modeFlags & RR_DoubleScan))
@@ -942,6 +943,96 @@ output_is_primary(output_t *output)
     return False;
 }
 
+/* Returns the index of the last value in an array < 0xffff */
+static int
+find_last_non_clamped(CARD16 array[], int size) {
+    int i;
+    for (i = size - 1; i > 0; i--) {
+        if (array[i] < 0xffff)
+	    return i;
+    }
+    return 0;
+}
+
+static void
+set_gamma_info(output_t *output)
+{
+    XRRCrtcGamma *gamma;
+    double i1, v1, i2, v2;
+    int size, middle, last_best, last_red, last_green, last_blue;
+    CARD16 *best_array;
+
+    if (!output->crtc_info)
+	return;
+
+    size = XRRGetCrtcGammaSize(dpy, output->crtc_info->crtc.xid);
+    if (!size) {
+	warning("Failed to get size of gamma for output %s\n", output->output.string);
+	return;
+    }
+
+    gamma = XRRGetCrtcGamma(dpy, output->crtc_info->crtc.xid);
+    if (!gamma) {
+	warning("Failed to get gamma for output %s\n", output->output.string);
+	return;
+    }
+
+    /*
+     * Here is a bit tricky because gamma is a whole curve for each
+     * color.  So, typically, we need to represent 3 * 256 values as 3 + 1
+     * values.  Therefore, we approximate the gamma curve (v) by supposing
+     * it always follows the way we set it: a power function (i^g)
+     * multiplied by a brightness (b).
+     * v = i^g * b
+     * so g = (ln(v) - ln(b))/ln(i)
+     * and b can be found using two points (v1,i1) and (v2, i2):
+     * b = e^((ln(v2)*ln(i1) - ln(v1)*ln(i2))/ln(i1/i2))
+     * For the best resolution, we select i2 at the highest place not
+     * clamped and i1 at i2/2. Note that if i2 = 1 (as in most normal
+     * cases), then b = v2.
+     */
+    last_red = find_last_non_clamped(gamma->red, size);
+    last_green = find_last_non_clamped(gamma->green, size);
+    last_blue = find_last_non_clamped(gamma->blue, size);
+    best_array = gamma->red;
+    last_best = last_red;
+    if (last_green > last_best) {
+	last_best = last_green;
+	best_array = gamma->green;
+    }
+    if (last_blue > last_best) {
+	last_best = last_blue;
+	best_array = gamma->blue;
+    }
+    if (last_best == 0)
+	last_best = 1;
+
+    middle = last_best / 2;
+    i1 = (double)(middle + 1) / size;
+    v1 = (double)(best_array[middle]) / 65535;
+    i2 = (double)(last_best + 1) / size;
+    v2 = (double)(best_array[last_best]) / 65535;
+    if (v2 < 0.0001) { /* The screen is black */
+	output->brightness = 0;
+	output->gamma.red = 1;
+	output->gamma.green = 1;
+	output->gamma.blue = 1;
+    } else {
+	if ((last_best + 1) == size)
+	    output->brightness = v2;
+	else
+	    output->brightness = exp((log(v2)*log(i1) - log(v1)*log(i2))/log(i1/i2));
+	output->gamma.red = log((double)(gamma->red[last_red / 2]) / output->brightness
+				/ 65535) / log((double)((last_red / 2) + 1) / size);
+	output->gamma.green = log((double)(gamma->green[last_green / 2]) / output->brightness
+				  / 65535) / log((double)((last_green / 2) + 1) / size);
+	output->gamma.blue = log((double)(gamma->blue[last_blue / 2]) / output->brightness
+				 / 65535) / log((double)((last_blue / 2) + 1) / size);
+    }
+
+    XRRFreeGamma(gamma);
+}
+
 static void
 set_output_info (output_t *output, RROutput xid, XRROutputInfo *output_info)
 {
@@ -1054,6 +1145,10 @@ set_output_info (output_t *output, RROutput xid, XRROutputInfo *output_info)
 	       rotation_name (output->rotation),
 	       reflection_name (output->rotation));
 
+    /* set gamma */
+    if (!(output->changes & changes_gamma))
+	    set_gamma_info(output);
+
     /* set transformation */
     if (!(output->changes & changes_transform))
     {
@@ -1096,9 +1191,7 @@ get_crtcs (void)
     for (c = 0; c < res->ncrtc; c++)
     {
 	XRRCrtcInfo *crtc_info = XRRGetCrtcInfo (dpy, res, res->crtcs[c]);
-#if RANDR_MAJOR > 1 || RANDR_MINOR >= 3
 	XRRCrtcTransformAttributes  *attr;
-#endif
 	XRRPanning  *panning_info = NULL;
 
 	if (has_1_3) {
@@ -1124,7 +1217,6 @@ get_crtcs (void)
 	    crtcs[c].y = 0;
 	    crtcs[c].rotation = RR_Rotate_0;
 	}
-#if RANDR_MAJOR > 1 || RANDR_MINOR >= 3
 	if (XRRGetCrtcTransform (dpy, res->crtcs[c], &attr) && attr) {
 	    set_transform (&crtcs[c].current_transform,
 			   &attr->currentTransform,
@@ -1134,7 +1226,6 @@ get_crtcs (void)
 	    XFree (attr);
 	}
 	else
-#endif
 	{
 	    init_transform (&crtcs[c].current_transform);
 	}
@@ -1223,24 +1314,30 @@ set_gamma(void)
 	    continue;
 	}
 
+	if(output->gamma.red == 0.0 && output->gamma.green == 0.0 && output->gamma.blue == 0.0)
+	    output->gamma.red = output->gamma.green = output->gamma.blue = 1.0;
+
 	for (i = 0; i < size; i++) {
-	    if (output->gamma.red == 1.0)
-		gamma->red[i] = i << 8;
+	    if (output->gamma.red == 1.0 && output->brightness == 1.0)
+		gamma->red[i] = (i << 8) + i;
 	    else
-		gamma->red[i] = (CARD16)(pow((double)i/(double)(size - 1),
-			    (double)output->gamma.red) * (double)(size - 1) * 256);
+		gamma->red[i] = dmin(pow((double)i/(double)(size - 1),
+					 output->gamma.red) * output->brightness,
+				     1.0) * 65535.0;
 
-	    if (output->gamma.green == 1.0)
-		gamma->green[i] = i << 8;
+	    if (output->gamma.green == 1.0 && output->brightness == 1.0)
+		gamma->green[i] = (i << 8) + i;
 	    else
-		gamma->green[i] = (CARD16)(pow((double)i/(double)(size - 1),
-			    (double)output->gamma.green) * (double)(size - 1) * 256);
+		gamma->green[i] = dmin(pow((double)i/(double)(size - 1),
+					   output->gamma.green) * output->brightness,
+				       1.0) * 65535.0;
 
-	    if (output->gamma.blue == 1.0)
-		gamma->blue[i] = i << 8;
+	    if (output->gamma.blue == 1.0 && output->brightness == 1.0)
+		gamma->blue[i] = (i << 8) + i;
 	    else
-		gamma->blue[i] = (CARD16)(pow((double)i/(double)(size - 1),
-			    (double)output->gamma.blue) * (double)(size - 1) * 256);
+		gamma->blue[i] = dmin(pow((double)i/(double)(size - 1),
+					  output->gamma.blue) * output->brightness,
+				      1.0) * 65535.0;
 	}
 
 	XRRSetCrtcGamma(dpy, crtc->crtc.xid, gamma);
@@ -1527,6 +1624,7 @@ static void
 get_outputs (void)
 {
     int		o;
+    output_t    *q;
     
     for (o = 0; o < res->noutput; o++)
     {
@@ -1565,6 +1663,7 @@ get_outputs (void)
 		}
 	    }
 	}
+	output->found = True;
 
 	/*
 	 * Automatic mode -- track connection state and enable/disable outputs
@@ -1595,6 +1694,14 @@ get_outputs (void)
 
 	set_output_info (output, res->outputs[o], output_info);
     }
+    for (q = outputs; q; q = q->next)
+    {
+	if (!q->found)
+	{
+	    fprintf(stderr, "warning: output %s not found; ignoring\n",
+		    q->output.string);
+	}
+    }
 }
 
 static void
@@ -1756,23 +1863,23 @@ set_positions (void)
 	    }
 	    
 	    switch (output->relation) {
-	    case left_of:
+	    case relation_left_of:
 		output->y = relation->y;
 		output->x = relation->x - mode_width (output->mode_info, output->rotation);
 		break;
-	    case right_of:
+	    case relation_right_of:
 		output->y = relation->y;
 		output->x = relation->x + mode_width (relation->mode_info, relation->rotation);
 		break;
-	    case above:
+	    case relation_above:
 		output->x = relation->x;
 		output->y = relation->y - mode_height (output->mode_info, output->rotation);
 		break;
-	    case below:
+	    case relation_below:
 		output->x = relation->x;
 		output->y = relation->y + mode_height (relation->mode_info, relation->rotation);
 		break;
-	    case same_as:
+	    case relation_same_as:
 		output->x = relation->x;
 		output->y = relation->y;
 	    }
@@ -1873,7 +1980,6 @@ set_screen_size (void)
     }
 }
     
-#endif
 
 static void
 disable_outputs (output_t *outputs)
@@ -1999,6 +2105,26 @@ pick_crtcs (void)
     }
 }
 
+static int
+check_strtol(char *s)
+{
+    char *endptr;
+    int result = strtol(s, &endptr, 10);
+    if (s == endptr)
+	usage();
+    return result;
+}
+
+static double
+check_strtod(char *s)
+{
+    char *endptr;
+    double result = strtod(s, &endptr);
+    if (s == endptr)
+	usage();
+    return result;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -2009,7 +2135,8 @@ main (int argc, char **argv)
     short		*rates;
     Status	status = RRSetConfigFailed;
     int		rot = -1;
-    int		query = 0;
+    int		query = False;
+    int		action_requested = False;
     Rotation	rotation, current_rotation, rotations;
     XEvent	event;
     XRRScreenChangeNotifyEvent *sce;    
@@ -2017,7 +2144,7 @@ main (int argc, char **argv)
     int 		i, j;
     SizeID	current_size;
     short	current_rate;
-    float    	rate = -1;
+    double    	rate = -1;
     int		size = -1;
     int		dirind = 0;
     Bool	setit = False;
@@ -2027,9 +2154,7 @@ main (int argc, char **argv)
     int		width = 0, height = 0;
     Bool    	have_pixel_size = False;
     int		ret = 0;
-#if HAS_RANDR_1_2
     output_t	*output = NULL;
-    policy_t	policy = clone;
     Bool    	setit_1_2 = False;
     Bool    	query_1_2 = False;
     Bool	modeit = False;
@@ -2037,10 +2162,8 @@ main (int argc, char **argv)
     Bool	query_1 = False;
     int		major, minor;
     Bool	current = False;
-#endif
 
     program_name = argv[0];
-    if (argc == 1) query = True;
     for (i = 1; i < argc; i++) {
 	if (!strcmp ("-display", argv[i]) || !strcmp ("-d", argv[i])) {
 	    if (++i>=argc) usage ();
@@ -2049,6 +2172,7 @@ main (int argc, char **argv)
 	}
 	if (!strcmp("-help", argv[i])) {
 	    usage();
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--verbose", argv[i])) {
@@ -2066,20 +2190,19 @@ main (int argc, char **argv)
 	}
 	if (!strcmp("--current", argv[i])) {
 	    current = True;
-	    /* if --current was the only arg, then query */
-	    if (argc == 2) query = True;
 	    continue;
 	}
 
 	if (!strcmp ("-s", argv[i]) || !strcmp ("--size", argv[i])) {
 	    if (++i>=argc) usage ();
-	    if (sscanf (argv[i], "%dx%d", &width, &height) == 2)
+	    if (sscanf (argv[i], "%dx%d", &width, &height) == 2) {
 		have_pixel_size = True;
-	    else {
-		size = atoi (argv[i]);
-		if (size < 0) usage();
-	    }
+	    } else {
+		size = check_strtol(argv[i]);
+                if (size < 0) usage();
+            }
 	    setit = True;
+	    action_requested = True;
 	    continue;
 	}
 
@@ -2088,38 +2211,39 @@ main (int argc, char **argv)
 	    !strcmp ("--refresh", argv[i]))
 	{
 	    if (++i>=argc) usage ();
-	    if (sscanf (argv[i], "%f", &rate) != 1)
-		usage ();
+	    rate = check_strtod(argv[i]);
 	    setit = True;
-#if HAS_RANDR_1_2
 	    if (output)
 	    {
 		output->refresh = rate;
 		output->changes |= changes_refresh;
 		setit_1_2 = True;
 	    }
-#endif
+	    action_requested = True;
 	    continue;
 	}
 
 	if (!strcmp ("-v", argv[i]) || !strcmp ("--version", argv[i])) {
 	    version = True;
+	    action_requested = True;
 	    continue;
 	}
 
 	if (!strcmp ("-x", argv[i])) {
 	    reflection |= RR_Reflect_X;
 	    setit = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("-y", argv[i])) {
 	    reflection |= RR_Reflect_Y;
 	    setit = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--screen", argv[i])) {
 	    if (++i>=argc) usage ();
-	    screen = atoi (argv[i]);
+	    screen = check_strtol(argv[i]);
 	    if (screen < 0) usage();
 	    continue;
 	}
@@ -2130,8 +2254,8 @@ main (int argc, char **argv)
 	if (!strcmp ("-o", argv[i]) || !strcmp ("--orientation", argv[i])) {
 	    char *endptr;
 	    if (++i>=argc) usage ();
-	    dirind = strtol(argv[i], &endptr, 0);
-	    if (*endptr != '\0') {
+	    dirind = strtol(argv[i], &endptr, 10);
+	    if (argv[i] == endptr) {
 		for (dirind = 0; dirind < 4; dirind++) {
 		    if (strcmp (direction[dirind], argv[i]) == 0) break;
 		}
@@ -2139,9 +2263,9 @@ main (int argc, char **argv)
 	    }
 	    rot = dirind;
 	    setit = True;
+	    action_requested = True;
 	    continue;
 	}
-#if HAS_RANDR_1_2
 	if (!strcmp ("--prop", argv[i]) ||
 	    !strcmp ("--props", argv[i]) ||
 	    !strcmp ("--madprops", argv[i]) ||
@@ -2149,6 +2273,7 @@ main (int argc, char **argv)
 	{
 	    query_1_2 = True;
 	    properties = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--output", argv[i])) {
@@ -2161,6 +2286,7 @@ main (int argc, char **argv)
 	    }
 	    
 	    setit_1_2 = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--crtc", argv[i])) {
@@ -2221,7 +2347,7 @@ main (int argc, char **argv)
 	if (!strcmp ("--left-of", argv[i])) {
 	    if (++i>=argc) usage ();
 	    if (!output) usage();
-	    output->relation = left_of;
+	    output->relation = relation_left_of;
 	    output->relative_to = argv[i];
 	    output->changes |= changes_relation;
 	    continue;
@@ -2229,7 +2355,7 @@ main (int argc, char **argv)
 	if (!strcmp ("--right-of", argv[i])) {
 	    if (++i>=argc) usage ();
 	    if (!output) usage();
-	    output->relation = right_of;
+	    output->relation = relation_right_of;
 	    output->relative_to = argv[i];
 	    output->changes |= changes_relation;
 	    continue;
@@ -2237,7 +2363,7 @@ main (int argc, char **argv)
 	if (!strcmp ("--above", argv[i])) {
 	    if (++i>=argc) usage ();
 	    if (!output) usage();
-	    output->relation = above;
+	    output->relation = relation_above;
 	    output->relative_to = argv[i];
 	    output->changes |= changes_relation;
 	    continue;
@@ -2245,7 +2371,7 @@ main (int argc, char **argv)
 	if (!strcmp ("--below", argv[i])) {
 	    if (++i>=argc) usage ();
 	    if (!output) usage();
-	    output->relation = below;
+	    output->relation = relation_below;
 	    output->relative_to = argv[i];
 	    output->changes |= changes_relation;
 	    continue;
@@ -2253,7 +2379,7 @@ main (int argc, char **argv)
 	if (!strcmp ("--same-as", argv[i])) {
 	    if (++i>=argc) usage ();
 	    if (!output) usage();
-	    output->relation = same_as;
+	    output->relation = relation_same_as;
 	    output->relative_to = argv[i];
 	    output->changes |= changes_relation;
 	    continue;
@@ -2298,6 +2424,15 @@ main (int argc, char **argv)
 	    setit_1_2 = True;
 	    continue;
 	}
+	if (!strcmp ("--brightness", argv[i])) {
+	    if (!output) usage();
+	    if (++i>=argc) usage();
+	    if (sscanf(argv[i], "%f", &output->brightness) != 1)
+		usage ();
+	    output->changes |= changes_gamma;
+	    setit_1_2 = True;
+	    continue;
+	}
 	if (!strcmp ("--primary", argv[i])) {
 	    if (!output) usage();
 	    output->changes |= changes_primary;
@@ -2328,6 +2463,7 @@ main (int argc, char **argv)
 	if (!strcmp ("--scale", argv[i]))
 	{
 	    double  sx, sy;
+	    if (!output) usage();
 	    if (++i>=argc) usage();
 	    if (sscanf (argv[i], "%lfx%lf", &sx, &sy) != 2)
 		usage ();
@@ -2347,6 +2483,7 @@ main (int argc, char **argv)
 	if (!strcmp ("--transform", argv[i])) {
 	    double  transform[3][3];
 	    int	    k, l;
+	    if (!output) usage();
 	    if (++i>=argc) usage ();
 	    init_transform (&output->transform);
 	    if (strcmp (argv[i], "none") != 0)
@@ -2382,6 +2519,7 @@ main (int argc, char **argv)
 			&fb_width, &fb_height) != 2)
 		usage ();
 	    setit_1_2 = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--fbmm", argv[i])) {
@@ -2390,26 +2528,20 @@ main (int argc, char **argv)
 			&fb_width_mm, &fb_height_mm) != 2)
 		usage ();
 	    setit_1_2 = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--dpi", argv[i])) {
+	    char *strtod_error;
 	    if (++i>=argc) usage ();
-	    if (sscanf (argv[i], "%f", &dpi) != 1)
+	    dpi = strtod(argv[i], &strtod_error);
+	    if (argv[i] == strtod_error)
 	    {
 		dpi = 0.0;
 		dpi_output = argv[i];
 	    }
 	    setit_1_2 = True;
-	    continue;
-	}
-	if (!strcmp ("--clone", argv[i])) {
-	    policy = clone;
-	    setit_1_2 = True;
-	    continue;
-	}
-	if (!strcmp ("--extend", argv[i])) {
-	    policy = extend;
-	    setit_1_2 = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--auto", argv[i])) {
@@ -2421,6 +2553,7 @@ main (int argc, char **argv)
 	    else
 		automatic = True;
 	    setit_1_2 = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--q12", argv[i]))
@@ -2436,25 +2569,24 @@ main (int argc, char **argv)
 	if (!strcmp ("--newmode", argv[i]))
 	{
 	    umode_t  *m = malloc (sizeof (umode_t));
-	    float   clock;
+	    double    clock;
 	    
 	    ++i;
 	    if (i + 9 >= argc) usage ();
 	    m->mode.name = argv[i];
 	    m->mode.nameLength = strlen (argv[i]);
 	    i++;
-	    if (sscanf (argv[i++], "%f", &clock) != 1)
-		usage ();
+	    clock = check_strtod(argv[i++]);
 	    m->mode.dotClock = clock * 1e6;
 
-	    if (sscanf (argv[i++], "%d", &m->mode.width) != 1) usage();
-	    if (sscanf (argv[i++], "%d", &m->mode.hSyncStart) != 1) usage();
-	    if (sscanf (argv[i++], "%d", &m->mode.hSyncEnd) != 1) usage();
-	    if (sscanf (argv[i++], "%d", &m->mode.hTotal) != 1) usage();
-	    if (sscanf (argv[i++], "%d", &m->mode.height) != 1) usage();
-	    if (sscanf (argv[i++], "%d", &m->mode.vSyncStart) != 1) usage();
-	    if (sscanf (argv[i++], "%d", &m->mode.vSyncEnd) != 1) usage();
-	    if (sscanf (argv[i++], "%d", &m->mode.vTotal) != 1) usage();
+	    m->mode.width = check_strtol(argv[i++]);
+	    m->mode.hSyncStart = check_strtol(argv[i++]);
+	    m->mode.hSyncEnd = check_strtol(argv[i++]);
+	    m->mode.hTotal = check_strtol(argv[i++]);
+	    m->mode.height = check_strtol(argv[i++]);
+	    m->mode.vSyncStart = check_strtol(argv[i++]);
+	    m->mode.vSyncEnd = check_strtol(argv[i++]);
+	    m->mode.vTotal = check_strtol(argv[i++]);
 	    m->mode.modeFlags = 0;
 	    while (i < argc) {
 		int f;
@@ -2472,6 +2604,7 @@ main (int argc, char **argv)
 	    m->action = umode_create;
 	    umodes = m;
 	    modeit = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--rmmode", argv[i]))
@@ -2484,6 +2617,7 @@ main (int argc, char **argv)
 	    m->next = umodes;
 	    umodes = m;
 	    modeit = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--addmode", argv[i]))
@@ -2498,6 +2632,7 @@ main (int argc, char **argv)
 	    m->next = umodes;
 	    umodes = m;
 	    modeit = True;
+	    action_requested = True;
 	    continue;
 	}
 	if (!strcmp ("--delmode", argv[i]))
@@ -2512,11 +2647,13 @@ main (int argc, char **argv)
 	    m->next = umodes;
 	    umodes = m;
 	    modeit = True;
+	    action_requested = True;
 	    continue;
 	}
-#endif
 	usage();
     }
+    if (!action_requested)
+	    query = True;
     if (verbose) 
     {
 	query = True;
@@ -2542,7 +2679,6 @@ main (int argc, char **argv)
 
     root = RootWindow (dpy, screen);
 
-#if HAS_RANDR_1_2
     if (!XRRQueryVersion (dpy, &major, &minor))
     {
 	fprintf (stderr, "RandR extension missing\n");
@@ -2877,6 +3013,11 @@ main (int argc, char **argv)
 		printf ("\tIdentifier: 0x%x\n", (int)output->output.xid);
 		printf ("\tTimestamp:  %d\n", (int)output_info->timestamp);
 		printf ("\tSubpixel:   %s\n", order[output_info->subpixel_order]);
+	        if (output->gamma.red != 0.0 && output->gamma.green != 0.0 && output->gamma.blue != 0.0) {
+		    printf ("\tGamma:      %#.2g:%#.2g:%#.2g\n",
+			    output->gamma.red, output->gamma.green, output->gamma.blue);
+		    printf ("\tBrightness: %#.2g\n", output->brightness);
+		}
 		printf ("\tClones:    ");
 		for (j = 0; j < output_info->nclone; j++)
 		{
@@ -3022,7 +3163,7 @@ main (int argc, char **argv)
 		    
 		    printf ("  %s (0x%x) %6.1fMHz",
 			    mode->name, (int)mode->id,
-			    (float)mode->dotClock / 1000000.0);
+			    (double)mode->dotClock / 1000000.0);
 		    for (f = 0; mode_flags[f].flag; f++)
 			if (mode->modeFlags & mode_flags[f].flag)
 			    printf (" %s", mode_flags[f].string);
@@ -3083,7 +3224,7 @@ main (int argc, char **argv)
 	    {
 		printf ("  %s (0x%x) %6.1fMHz\n",
 			mode->name, (int)mode->id,
-			(float)mode->dotClock / 1000000.0);
+			(double)mode->dotClock / 1000000.0);
 		printf ("        h: width  %4d start %4d end %4d total %4d skew %4d clock %6.1fKHz\n",
 			mode->width, mode->hSyncStart, mode->hSyncEnd,
 			mode->hTotal, mode->hSkew, mode_hsync (mode) / 1000);
@@ -3094,7 +3235,6 @@ main (int argc, char **argv)
 	}
 	exit (0);
     }
-#endif
     
     sc = XRRGetScreenInfo (dpy, root);
 
